# HList.tcl --
#
# This file defines the default bindings for Tix Hierarchical Listbox widgets.
#
# Copyright (c) 1995 Ioi K Lam
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
#--------------------------------------------------------------------------
# tkPriv elements used in this file:
#
# afterId -		Token returned by "after" for autoscanning.
# fakeRelease -		Cancel the ButtonRelease-1 after the user double click
#--------------------------------------------------------------------------
#
# ToDo: shift-up,down keys
proc tixHListBind {} {
    bind TixHList <ButtonPress-1> {
	tixHList::Button-1 %W %y
    }
    bind TixHList <Shift-ButtonPress-1> {
	tixHList::Shift-Button-1 %W %y
    }
    bind TixHList <Control-ButtonRelease-1> {;}
    bind TixHList <ButtonRelease-1> {
	tkCancelRepeat
	tixHList::ButtonRelease-1 %W %y
    }
    bind TixHList <B1-Motion> {
	tixHList::B1-Motion %W %y
    }
    bind TixHList <Control-B1-Motion> {
	;
    }
    bind TixHList <Control-ButtonPress-1> {
	tixHList::Ctrl-Button1 %W %y
    }
    bind TixHList <Control-Double-ButtonPress-1> {
	tixHList::Ctrl-Button1 %W %y
    }
    bind TixHList <Double-ButtonPress-1> {
	tixHList::Double1 %W %y
    }
    bind TixHList <Enter> {
	tixHList::Enter %W
    }
    bind TixHList  <B1-Leave> {
	tixHList::AutoScan %W %x %y
    }
    bind TixHList <B1-Enter> {
	tkCancelRepeat
    }
    bind TixHList <Up> {
	tixHList:UpDown %W prev 0
    }
    bind TixHList <Down> {
	tixHList:UpDown %W next 0
    }
    bind TixHList <Shift-Up> {
	tixHList:UpDown %W prev 1
    }
    bind TixHList <Shift-Down> {
	tixHList:UpDown %W next 1
    }
    bind TixHList <Left> {
	tixHList:LeftRight %W left
    }
    bind TixHList <Right> {
	tixHList:LeftRight %W right
    }
    bind TixHList <Prior> {
	%W yview scroll -1 pages
    }
    bind TixHList <Next> {
	%W yview scroll 1 pages
    }
    bind TixHList <Return> {
	tixHList:Keyboard-Activate %W 
    }
    bind TixHList <space> {
	tixHList:Keyboard-Browse %W 
    }
}

#----------------------------------------------------------------------
# Key bindings
#----------------------------------------------------------------------
proc tixHList:Keyboard-Activate {w} {
    set anchor [$w info anchor]

    if {$anchor == ""} {
	return
    }

    if {[$w cget -selectmode] == "single"} {
	$w select clear
	$w select set $anchor
    }

    set command [$w cget -command]
    if {$command != {}} {
	eval $command [list $anchor]
    }
}

proc tixHList:Keyboard-Browse {w} {
    set anchor [$w info anchor]

    if {$anchor == ""} {
	return
    }

    if {[$w cget -selectmode] == "single"} {
	$w select clear
	$w select set $anchor
    }

    set browsecmd [$w cget -browsecmd]
    if {$browsecmd != {}} {
	eval $browsecmd [list $anchor]
    }
}

proc tixHList:LeftRight {w spec} {

    set anchor [$w info anchor]
    if {$anchor == ""} {
	set anchor [lindex [$w info children] 0]
    }

    set done 0
    set ent $anchor
    while {!$done} {
	set e $ent
	if {$spec == "left"} {
	    set ent [$w info parent $e]

	    if {$ent == {} || [$w entrycget $ent -state] == "disabled"} {
		set ent [$w info prev $e]
	    }
	} else {
	    set ent [lindex [$w info children $e] 0]

	    if {$ent == {} || [$w entrycget $ent -state] == "disabled"} {
		set ent [$w info next $e]
	    }
	}

	if {$ent == {}} {
	    break
	}
 	if {[$w entrycget $ent -state] != "disabled"} {
	    break
	}
   }

   if {$ent == {}} {
	if {$spec == "left"} {
	    $w xview scroll -1 unit
	} else {
	    $w xview scroll 1 unit
	}
	return
    }

    $w anchor set $ent
    $w see $ent

    if {[$w cget -selectmode] != "single"} {
	$w select clear
	$w selection set $ent

	set browsecmd [$w cget -browsecmd]
	if {$browsecmd != {}} {
	    eval $browsecmd [list $ent]
	}
    }
}

proc tixHList:UpDown {w spec shift} {

    set anchor [$w info anchor]
    if {$anchor == ""} {
	set anchor [lindex [$w info children] 0]
    }

    set done 0
    set ent $anchor
    while {!$done} {
	set ent [$w info $spec $ent]
	if {$ent == {}} {
	    break
	}
	if {[$w entrycget $ent -state] != "disabled"} {
	    break
	}
    }

    if {$ent == {}} {
	if {$spec == "prev"} {
	    $w yview scroll -1 unit
	} else {
	    $w yview scroll 1 unit
	}
	return
    }

    $w anchor set $ent
    $w see $ent

    if {[$w cget -selectmode] != "single"} {
	$w select clear
	$w selection set $ent

	set browsecmd [$w cget -browsecmd]
	if {$browsecmd != {}} {
	    eval $browsecmd [list $ent]
	}
    }
}

#----------------------------------------------------------------------
# Mouse bindings
#----------------------------------------------------------------------

proc tixHList::Enter {w} {
    # Dummy function for auto-loading this file
    #
}

proc tixHList::GetNearest {w y} {
    set ent [$w nearest $y]

    if {$ent != {}} {
	if {[$w entrycget $ent -state] != "disabled"} {
	    return $ent

	}
    }
    return {}
}

proc tixHList::Button-1 {w y} {
    if [$w cget -takefocus] {
	focus $w
    }

    set ent [tixHList::GetNearest $w $y]

    if {$ent != ""} {
	case [$w cget -selectmode] {
	    {single} {
		$w anchor set $ent
	    }
	    {browse} {
		$w anchor set $ent
		$w select clear
		$w select set $ent
	    }
	    {multiple} {
		$w select clear
		$w anchor set $ent
		$w select set $ent
	    }
	    {extended} {
		$w anchor set $ent
		$w select clear
		$w select set $ent
	    }
	}

	if {[$w cget -selectmode] != "single"} {
	    set browsecmd [$w cget -browsecmd]
	    if {$browsecmd != {}} {
		eval $browsecmd [list $ent]
	    }
	}
    }
}

proc tixHList::Shift-Button-1 {w y} {
    set to [tixHList::GetNearest $w $y]

    if {$to != ""} {
	case [$w cget -selectmode] {
	    {multiple extended} {
		set from [$w info anchor]
		if {$from == {}} {
		    $w anchor set $to
		    $w select clear
		    $w select set $to
		} else {
		    $w select clear
		    $w select set $from $to
		}
	    }
	}
    }
}

proc tixHList::ButtonRelease-1 {w y} {
    global tkPriv

    if {[info exists tkPriv(fakeRelease)]} {
	if {$tkPriv(fakeRelease) == $w} {
	    catch {unset tkPriv(fakeRelease)}
	    return
	} else {
	    catch {unset tkPriv(fakeRelease)}
	}
    }

    set ent [tixHList::GetNearest $w $y]
    if {$ent == ""} {
	set ent [$w info anchor]
    }


    if {$ent != ""} {
	case [$w cget -selectmode] {
	    {single browse} {
		$w anchor set $ent
		$w select clear
		$w select set $ent
	    }
	    {multiple} {
		$w select set $ent
	    }
	    {extended} {
		$w select set $ent
	    }
	}

	set browsecmd [$w cget -browsecmd]
	if {$browsecmd != {}} {
	    eval $browsecmd [list $ent]
	}
    }
}

proc tixHList::Double1 {w y} {
    global tkPriv
    set ent [tixHList::GetNearest $w $y]

    if {$ent != ""} {
	if {[$w info anchor] == {}} {
	    $w anchor set $ent
	}
	$w select set $ent
	set command [$w cget -command]
	if {$command != {}} {
	    eval $command [list $ent]
	}
    }

    set tkPriv(fakeRelease) $w
}

proc tixHList::Ctrl-Button1 {w y} {
    set ent [tixHList::GetNearest $w $y]

    if {$ent != ""} {
	case [$w cget -selectmode] {
	    {extended} {
		if {[$w info anchor] == {}} {
		    $w anchor set $ent
		}

		if [$w select includes $ent] {
		    $w select clear $ent
		} else {
		    $w select set $ent
		}
		
		set browsecmd [$w cget -browsecmd]
		if {$browsecmd != {}} {
		    eval $browsecmd [list $ent]
		}
	    }
	}
    }
}

# tixHList::AutoScan --
# This procedure is invoked when the mouse leaves an entry window
# with button 1 down.  It scrolls the window up, down, left, or
# right, depending on where the mouse left the window, and reschedules
# itself as an "after" command so that the window continues to scroll until
# the mouse moves back into the window or the mouse button is released.
#
# Arguments:
# w -		The entry window.
# x -		The x-coordinate of the mouse when it left the window.
# y -		The y-coordinate of the mouse when it left the window.

proc tixHList::AutoScan {w x y} {
    global tkPriv

    if {$y >= [winfo height $w]} {
	$w yview scroll 1 units
    } elseif {$y < 0} {
	$w yview scroll -1 units
    } elseif {$x >= [winfo width $w]} {
	$w xview scroll 2 units
    } elseif {$x < 0} {
	$w xview scroll -2 units
    } else {
	return
    }

    tixHList::B1-Motion $w $y

    set tkPriv(afterId) [after 50 tixHList::AutoScan $w $x $y]
}

# ToDo: The multiple selection is not very efficient when there
#	are a large number of entries. Should only select/de-select
#	those entries who are affected.
# 
#
proc tixHList::B1-Motion {w y} {
    set ent [tixHList::GetNearest $w $y]

    if {$ent != ""} {
	case [$w cget -selectmode] {
	    {single} {
		$w anchor set $ent
	    }
	    {browse} {
		$w select clear
		$w select set $ent
		$w anchor set $ent

		set browsecmd [$w cget -browsecmd]
		if {$browsecmd != {}} {
		    eval $browsecmd [list $ent]
		}
	    }
	    {multiple extended} {
		if {[$w info anchor] == {}} {
		    $w anchor set $ent
		    $w select clear
		    $w select set $ent
		} else {
		    set from [$w info anchor]
		    set to $ent
		    $w select clear

		    $w select set $from $to
		}
	    }
	}
    }
}
