# tixMultiList --
#
#	A listbox that supporu multiple columns
#
# Author's note
#
#
tixWidgetClass tixMultiList {
    -classname TixMultiList
    -superclass tixScrolledWidget
    -method {
	insert setcolumn
    }
    -flag {
	-columns -rows -widths
    }
    -static {
	-columns
    }
    -configspec {
	{-columns columns Columns 1}
	{-rows rows Rows {}}
	{-widths widths Widths {}}
    }
}

proc tixMultiList::InitWidgetRec {w} {
    upvar #0 $w data

    if {$data(-columns) < 1} {
	error "Wrong # of columns \"$data(-columns)\". Must be at least 1"
    }

    set data(dx) 0

}

proc tixMultiList::ConstructWidget {w} {
    upvar #0 $w data

    set data(pw:f1) \
	[frame $w.f1]
    set data(pw:f2) \
	[frame $w.f2 -bd 0]
    set data(pw:f3) \
	[frame $w.f2.f3 -bd 0]
    set data(w:hsb) \
	[scrollbar $w.hsb -orient horizontal]
    set data(w:vsb) \
	[scrollbar $w.vsb -orient vertical ]
    set data(pw:client) $data(pw:f1)

    for {set i 0} {$i < $data(-columns)} {incr i} {
	set data(w:list$i) [listbox $data(pw:f3).list$i \
			    -yscrollcommand "tixMultiList::LbScrollY $w"]
	if {$data(-rows) != {}} {
	    $data(w:list$i) config -height $data(-rows)
	}
	set width [lindex $data(-widths) $i]
	if {$width != {}} {
	    $data(w:list$i) config -width $width
	}
	pack $data(w:list$i) -side left -fill both
    }
    pack $data(pw:f2) -in $data(pw:f1) -expand yes -fill both
    pack $data(pw:f3) -expand yes -fill both
}


proc tixMultiList::SetBindings {w} {
    upvar #0 $w data

    tixChainMethod $w SetBindings

    $data(w:hsb) config -command "tixMultiList::ScrollX $w"
    $data(w:vsb) config -command "tixMultiList::ScrollY $w"

    tixManageGeometry $data(pw:f3) "tixMultiList::WindowGeomProc $w"
}

#----------------------------------------------------------------------
#		Public Methods
#----------------------------------------------------------------------
proc tixMultiList::insert {w index text} {
    upvar #0 $w data

    for {set i 0} {$i < $data(-columns)} {incr i} {
	$data(w:list$i) insert $index [lindex $text $i]
    }
}



#----------------------------------------------------------------------
#		Private Methods
#----------------------------------------------------------------------
proc tixMultiList::LbScrollY {w first last} {
    upvar #0 $w data

    $data(w:vsb) set $first $last
}

# This guy just keeps asking for a same size as the w:window 
#
proc tixMultiList::WindowGeomProc {w args} {
    if {![winfo exists $w]} {
	return
    }

    upvar #0 $w data

    set rw [winfo reqwidth  $data(pw:f3)]
    set rh [winfo reqheight $data(pw:f3)]

    tixGeometryRequest $data(pw:f2) $rw $rh
}

proc tixMultiList::ScrollX {w args} {
    upvar #0 $w data

    set bd \
       [expr [$data(pw:f1) cget -bd] + [$data(pw:f1) cget -highlightthickness]]
    set fw [expr [winfo width  $data(pw:f1)] - 2*$bd]
    set ww [winfo reqwidth  $data(pw:f3)]

    set data(dx) \
	[eval tixMultiList::Scroll $w $ww $fw $data(dx) $args]

    tixWidgetDoWhenIdle tixMultiList::PlaceWindow $w
}

proc tixMultiList::Scroll {w total window first args} {
    case [lindex $args 0] {
	"scroll" {
	    set amt  [lindex $args 1]
	    set unit [lindex $args 2]

	    case $unit {
		"units" {
		    set first [expr $first + $amt * $window / 16]
		}
		"pages" {
		    set first [expr $first + $amt * $window]
		}
	    }
	}
	"moveto" {
	    set to [lindex $args 1]
	    set first [expr int($to * $total)]
	}
    }

    if {[expr $first + $window] > $total} {
	set first [expr $total - $window]
    }
    if {$first < 0} {
	set first 0
    }

    return $first
}

proc tixMultiList::ScrollY {w args} {
    upvar #0 $w data

    for {set i 0} {$i < $data(-columns)} {incr i} {
	eval $data(w:list$i) yview $args
    }
}

proc tixMultiList::PlaceWindow {w} {
    upvar #0 $w data

    set bd \
       [expr [$data(pw:f1) cget -bd] + [$data(pw:f1) cget -highlightthickness]]
    set fw [expr [winfo width  $data(pw:f1)] - 2*$bd]
    set fh [expr [winfo height $data(pw:f1)] - 2*$bd]
    set ww [winfo reqwidth  $data(pw:f3)]

    tixMapWindow $data(pw:f3)
    tixMoveResizeWindow $data(pw:f3) -$data(dx)  0 $ww $fh

    set first [expr $data(dx).0 / $ww.0]
    set last  [expr $first + ($fw.0 / $ww.0)]
    $data(w:hsb) set $first $last
}

#----------------------------------------------------------------------
# virtual functions to query the client window's scroll requirement
#
# When this function is called, the scrolled window is going to be
# mapped, if it is still unmapped. Also, it is going to change its
# size. Therefore, it is a good time to check whether the pw:f3 needs
# to be re-positioned due to the new parent window size.
#----------------------------------------------------------------------
proc tixMultiList::GeometryInfo {w mW mH} {
    upvar #0 $w data

    set bd \
       [expr [$data(pw:f1) cget -bd] + [$data(pw:f1) cget -highlightthickness]]
    set fw [expr $mW -2*$bd]
    set ww [winfo reqwidth  $data(pw:f3)]

    # Calculate the X info
    #
    if {$fw >= $ww} {
	if {$data(dx) > 0} {
	    set data(dx) 0
	}
	set xinfo [list 0.0 1.0]
    } else {
	set maxdx [expr $ww - $fw]
	if {$data(dx) > $maxdx} {
	    set data(dx) $maxdx
	}
	set first [expr $data(dx).0 / $ww.0]
	set last  [expr $first + ($fw.0 / $ww.0)]
	set xinfo [list $first $last]
    }

    # Calculate the Y info
    #
    set yinfo [lindex [$data(w:list0) geometryinfo $mW $mH] 1]

    tixWidgetDoWhenIdle tixMultiList::PlaceWindow $w
 
    return [list $xinfo $yinfo]
}
